import datetime

from betterproto2_compiler.lib.google.protobuf import Duration as VanillaDuration


class Duration(VanillaDuration):
    @classmethod
    def from_timedelta(
        cls, delta: datetime.timedelta, *, _1_microsecond: datetime.timedelta = datetime.timedelta(microseconds=1)
    ) -> "Duration":
        total_ms = delta // _1_microsecond
        seconds = int(total_ms / 1e6)
        nanos = int((total_ms % 1e6) * 1e3)
        return cls(seconds, nanos)

    def to_timedelta(self) -> datetime.timedelta:
        return datetime.timedelta(seconds=self.seconds, microseconds=self.nanos / 1e3)

    @staticmethod
    def delta_to_json(delta: datetime.timedelta) -> str:
        parts = str(delta.total_seconds()).split(".")
        if len(parts) > 1:
            while len(parts[1]) not in (3, 6, 9):
                parts[1] = f"{parts[1]}0"
        return f"{'.'.join(parts)}s"



import typing

import betterproto2

from betterproto2_compiler.lib.google.protobuf import Any as VanillaAny

default_message_pool = betterproto2.MessagePool()  # Only for typing purpose


class Any(VanillaAny):
    def pack(self, message: betterproto2.Message, message_pool: "betterproto2.MessagePool | None" = None) -> None:
        """
        Pack the given message in the `Any` object.

        The message type must be registered in the message pool, which is done automatically when the module defining
        the message type is imported.
        """
        message_pool = message_pool or default_message_pool

        self.type_url = message_pool.type_to_url[type(message)]
        self.value = bytes(message)

    def unpack(self, message_pool: "betterproto2.MessagePool | None" = None) -> betterproto2.Message | None:
        """
        Return the message packed inside the `Any` object.

        The target message type must be registered in the message pool, which is done automatically when the module
        defining the message type is imported.
        """
        if not self.type_url:
            return None

        message_pool = message_pool or default_message_pool

        try:
            message_type = message_pool.url_to_type[self.type_url]
        except KeyError:
            raise TypeError(f"Can't unpack unregistered type: {self.type_url}")

        return message_type().parse(self.value)

    def to_dict(self, **kwargs) -> dict[str, typing.Any]:
        # TODO allow passing a message pool to `to_dict`
        output: dict[str, typing.Any] = {"@type": self.type_url}

        value = self.unpack()

        if value is None:
            return output

        if type(value).to_dict == betterproto2.Message.to_dict:
            output.update(value.to_dict(**kwargs))
        else:
            output["value"] = value.to_dict(**kwargs)

        return output



import datetime

from betterproto2_compiler.lib.google.protobuf import Timestamp as VanillaTimestamp


class Timestamp(VanillaTimestamp):
    @classmethod
    def from_datetime(cls, dt: datetime.datetime) -> "Timestamp":
        # manual epoch offset calulation to avoid rounding errors,
        # to support negative timestamps (before 1970) and skirt
        # around datetime bugs (apparently 0 isn't a year in [0, 9999]??)
        offset = dt - datetime.datetime(1970, 1, 1, tzinfo=datetime.timezone.utc)
        # below is the same as timedelta.total_seconds() but without dividing by 1e6
        # so we end up with microseconds as integers instead of seconds as float
        offset_us = (offset.days * 24 * 60 * 60 + offset.seconds) * 10**6 + offset.microseconds
        seconds, us = divmod(offset_us, 10**6)
        return cls(seconds, us * 1000)

    def to_datetime(self) -> datetime.datetime:
        # datetime.fromtimestamp() expects a timestamp in seconds, not microseconds
        # if we pass it as a floating point number, we will run into rounding errors
        # see also #407
        offset = datetime.timedelta(seconds=self.seconds, microseconds=self.nanos // 1000)
        return datetime.datetime(1970, 1, 1, tzinfo=datetime.timezone.utc) + offset

    @staticmethod
    def timestamp_to_json(dt: datetime.datetime) -> str:
        nanos = dt.microsecond * 1e3
        if dt.tzinfo is not None:
            # change timezone aware datetime objects to utc
            dt = dt.astimezone(datetime.timezone.utc)
        copy = dt.replace(microsecond=0, tzinfo=None)
        result = copy.isoformat()
        if (nanos % 1e9) == 0:
            # If there are 0 fractional digits, the fractional
            # point '.' should be omitted when serializing.
            return f"{result}Z"
        if (nanos % 1e6) == 0:
            # Serialize 3 fractional digits.
            return f"{result}.{int(nanos // 1e6):03d}Z"
        if (nanos % 1e3) == 0:
            # Serialize 6 fractional digits.
            return f"{result}.{int(nanos // 1e3):06d}Z"
        # Serialize 9 fractional digits.
        return f"{result}.{nanos:09d}"



import re
from collections import defaultdict
from collections.abc import Iterator
from dataclasses import (
    dataclass,
    field,
)


@dataclass
class ModuleValidator:
    line_iterator: Iterator[str]
    line_number: int = field(init=False, default=0)

    collisions: dict[str, list[tuple[int, str]]] = field(init=False, default_factory=lambda: defaultdict(list))

    def add_import(self, imp: str, number: int, full_line: str):
        """
        Adds an import to be tracked.
        """
        self.collisions[imp].append((number, full_line))

    def process_import(self, imp: str):
        """
        Filters out the import to its actual value.
        """
        if " as " in imp:
            imp = imp[imp.index(" as ") + 4 :]

        imp = imp.strip()
        assert " " not in imp, imp
        return imp

    def evaluate_multiline_import(self, line: str):
        """
        Evaluates a multiline import from a starting line
        """
        # Filter the first line and remove anything before the import statement.
        full_line = line
        line = line.split("import", 1)[1]
        if "(" in line:
            conditional = lambda line: ")" not in line
        else:
            conditional = lambda line: "\\" in line

        # Remove open parenthesis if it exists.
        if "(" in line:
            line = line[line.index("(") + 1 :]

        # Choose the conditional based on how multiline imports are formatted.
        while conditional(line):
            # Split the line by commas
            imports = line.split(",")

            for imp in imports:
                # Add the import to the namespace
                imp = self.process_import(imp)
                if imp:
                    self.add_import(imp, self.line_number, full_line)
            # Get the next line
            full_line = line = next(self.line_iterator)
            # Increment the line number
            self.line_number += 1

        # validate the last line
        if ")" in line:
            line = line[: line.index(")")]
        imports = line.split(",")
        for imp in imports:
            imp = self.process_import(imp)
            if imp:
                self.add_import(imp, self.line_number, full_line)

    def evaluate_import(self, line: str):
        """
        Extracts an import from a line.
        """
        whole_line = line
        line = line[line.index("import") + 6 :]
        values = line.split(",")
        for v in values:
            self.add_import(self.process_import(v), self.line_number, whole_line)

    def next(self):
        """
        Evaluate each line for names in the module.
        """
        line = next(self.line_iterator)

        # Skip lines with indentation or comments
        if (
            # Skip indents and whitespace.
            line.startswith(" ")
            or line == "\n"
            or line.startswith("\t")
            or
            # Skip comments
            line.startswith("#")
            or
            # Skip  decorators
            line.startswith("@")
        ):
            self.line_number += 1
            return

        # Skip docstrings.
        if line.startswith('"""') or line.startswith("'''"):
            quote = line[0] * 3
            line = line[3:]
            while quote not in line:
                line = next(self.line_iterator)
            self.line_number += 1
            return

        # Evaluate Imports.
        if line.startswith("from ") or line.startswith("import "):
            if "(" in line or "\\" in line:
                self.evaluate_multiline_import(line)
            else:
                self.evaluate_import(line)

        # Evaluate Classes.
        elif match := re.search(r"^class (\w+)", line):
            class_name = match.group(1)
            if class_name:
                self.add_import(class_name, self.line_number, line)

        # Evaluate Functions.
        elif match := re.search(r"^def (\w+)", line):
            function_name = match.group(1)
            if function_name:
                self.add_import(function_name, self.line_number, line)

        # Evaluate direct assignments.
        elif match := re.search(r"^(\w+)\s*=", line):
            assignment = match.group(1)
            if assignment:
                self.add_import(assignment, self.line_number, line)

        self.line_number += 1

    def validate(self) -> bool:
        """
        Run Validation.
        """
        try:
            while True:
                self.next()
        except StopIteration:
            pass

        # Filter collisions for those with more than one value.
        self.collisions = {k: v for k, v in self.collisions.items() if len(v) > 1}

        # Return True if no collisions are found.
        return not bool(self.collisions)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: google/protobuf/compiler/plugin.proto
# plugin: python-betterproto2
# This file has been @generated

__all__ = (
    "CodeGeneratorResponseFeature",
    "CodeGeneratorRequest",
    "CodeGeneratorResponse",
    "CodeGeneratorResponseFile",
    "Version",
)


from dataclasses import dataclass
from typing import (
    List,
    Optional,
)

import betterproto2

from ....message_pool import default_message_pool

betterproto2.check_compiler_version("0.2.0")


class CodeGeneratorResponseFeature(betterproto2.Enum):
    """
    Sync with code_generator.h.
    """

    FEATURE_NONE = 0

    FEATURE_PROTO3_OPTIONAL = 1

    FEATURE_SUPPORTS_EDITIONS = 2


@dataclass(eq=False, repr=False)
class CodeGeneratorRequest(betterproto2.Message):
    """
    An encoded CodeGeneratorRequest is written to the plugin's stdin.
    """

    file_to_generate: "List[str]" = betterproto2.field(1, betterproto2.TYPE_STRING, repeated=True)
    """
    The .proto files that were explicitly listed on the command-line.  The
    code generator should generate code only for these files.  Each file's
    descriptor will be included in proto_file, below.
    """

    parameter: "str" = betterproto2.field(2, betterproto2.TYPE_STRING)
    """
    The generator parameter passed on the command-line.
    """

    proto_file: "List[__protobuf__.FileDescriptorProto]" = betterproto2.field(
        15, betterproto2.TYPE_MESSAGE, repeated=True
    )
    """
    FileDescriptorProtos for all files in files_to_generate and everything
    they import.  The files will appear in topological order, so each file
    appears before any file that imports it.

    Note: the files listed in files_to_generate will include runtime-retention
    options only, but all other files will include source-retention options.
    The source_file_descriptors field below is available in case you need
    source-retention options for files_to_generate.

    protoc guarantees that all proto_files will be written after
    the fields above, even though this is not technically guaranteed by the
    protobuf wire format.  This theoretically could allow a plugin to stream
    in the FileDescriptorProtos and handle them one by one rather than read
    the entire set into memory at once.  However, as of this writing, this
    is not similarly optimized on protoc's end -- it will store all fields in
    memory at once before sending them to the plugin.

    Type names of fields and extensions in the FileDescriptorProto are always
    fully qualified.
    """

    source_file_descriptors: "List[__protobuf__.FileDescriptorProto]" = betterproto2.field(
        17, betterproto2.TYPE_MESSAGE, repeated=True
    )
    """
    File descriptors with all options, including source-retention options.
    These descriptors are only provided for the files listed in
    files_to_generate.
    """

    compiler_version: "Optional[Version]" = betterproto2.field(3, betterproto2.TYPE_MESSAGE, optional=True)
    """
    The version number of protocol compiler.
    """


default_message_pool.register_message("google.protobuf.compiler", "CodeGeneratorRequest", CodeGeneratorRequest)


@dataclass(eq=False, repr=False)
class CodeGeneratorResponse(betterproto2.Message):
    """
    The plugin writes an encoded CodeGeneratorResponse to stdout.
    """

    error: "str" = betterproto2.field(1, betterproto2.TYPE_STRING)
    """
    Error message.  If non-empty, code generation failed.  The plugin process
    should exit with status code zero even if it reports an error in this way.

    This should be used to indicate errors in .proto files which prevent the
    code generator from generating correct code.  Errors which indicate a
    problem in protoc itself -- such as the input CodeGeneratorRequest being
    unparseable -- should be reported by writing a message to stderr and
    exiting with a non-zero status code.
    """

    supported_features: "int" = betterproto2.field(2, betterproto2.TYPE_UINT64)
    """
    A bitmask of supported features that the code generator supports.
    This is a bitwise "or" of values from the Feature enum.
    """

    file: "List[CodeGeneratorResponseFile]" = betterproto2.field(15, betterproto2.TYPE_MESSAGE, repeated=True)


default_message_pool.register_message("google.protobuf.compiler", "CodeGeneratorResponse", CodeGeneratorResponse)


@dataclass(eq=False, repr=False)
class CodeGeneratorResponseFile(betterproto2.Message):
    """
    Represents a single generated file.
    """

    name: "str" = betterproto2.field(1, betterproto2.TYPE_STRING)
    """
    The file name, relative to the output directory.  The name must not
    contain "." or ".." components and must be relative, not be absolute (so,
    the file cannot lie outside the output directory).  "/" must be used as
    the path separator, not "\".

    If the name is omitted, the content will be appended to the previous
    file.  This allows the generator to break large files into small chunks,
    and allows the generated text to be streamed back to protoc so that large
    files need not reside completely in memory at one time.  Note that as of
    this writing protoc does not optimize for this -- it will read the entire
    CodeGeneratorResponse before writing files to disk.
    """

    insertion_point: "str" = betterproto2.field(2, betterproto2.TYPE_STRING)
    """
    If non-empty, indicates that the named file should already exist, and the
    content here is to be inserted into that file at a defined insertion
    point.  This feature allows a code generator to extend the output
    produced by another code generator.  The original generator may provide
    insertion points by placing special annotations in the file that look
    like:
      @@protoc_insertion_point(NAME)
    The annotation can have arbitrary text before and after it on the line,
    which allows it to be placed in a comment.  NAME should be replaced with
    an identifier naming the point -- this is what other generators will use
    as the insertion_point.  Code inserted at this point will be placed
    immediately above the line containing the insertion point (thus multiple
    insertions to the same point will come out in the order they were added).
    The double-@ is intended to make it unlikely that the generated code
    could contain things that look like insertion points by accident.

    For example, the C++ code generator places the following line in the
    .pb.h files that it generates:
      // @@protoc_insertion_point(namespace_scope)
    This line appears within the scope of the file's package namespace, but
    outside of any particular class.  Another plugin can then specify the
    insertion_point "namespace_scope" to generate additional classes or
    other declarations that should be placed in this scope.

    Note that if the line containing the insertion point begins with
    whitespace, the same whitespace will be added to every line of the
    inserted text.  This is useful for languages like Python, where
    indentation matters.  In these languages, the insertion point comment
    should be indented the same amount as any inserted code will need to be
    in order to work correctly in that context.

    The code generator that generates the initial file and the one which
    inserts into it must both run as part of a single invocation of protoc.
    Code generators are executed in the order in which they appear on the
    command line.

    If |insertion_point| is present, |name| must also be present.
    """

    content: "str" = betterproto2.field(15, betterproto2.TYPE_STRING)
    """
    The file contents.
    """

    generated_code_info: "Optional[__protobuf__.GeneratedCodeInfo]" = betterproto2.field(
        16, betterproto2.TYPE_MESSAGE, optional=True
    )
    """
    Information describing the file content being inserted. If an insertion
    point is used, this information will be appropriately offset and inserted
    into the code generation metadata for the generated files.
    """


default_message_pool.register_message(
    "google.protobuf.compiler", "CodeGeneratorResponse.File", CodeGeneratorResponseFile
)


@dataclass(eq=False, repr=False)
class Version(betterproto2.Message):
    """
    The version number of protocol compiler.
    """

    major: "int" = betterproto2.field(1, betterproto2.TYPE_INT32)

    minor: "int" = betterproto2.field(2, betterproto2.TYPE_INT32)

    patch: "int" = betterproto2.field(3, betterproto2.TYPE_INT32)

    suffix: "str" = betterproto2.field(4, betterproto2.TYPE_STRING)
    """
    A suffix for alpha, beta or rc release, e.g., "alpha-1", "rc2". It should
    be empty for mainline stable releases.
    """


default_message_pool.register_message("google.protobuf.compiler", "Version", Version)


from ... import protobuf as __protobuf__



from __future__ import annotations

import os
from typing import (
    TYPE_CHECKING,
)

from betterproto2_compiler.lib.google import protobuf as google_protobuf
from betterproto2_compiler.settings import Settings

from ..casing import safe_snake_case
from .naming import pythonize_class_name

if TYPE_CHECKING:
    from ..plugin.models import PluginRequestCompiler

WRAPPER_TYPES: dict[str, type] = {
    ".google.protobuf.DoubleValue": google_protobuf.DoubleValue,
    ".google.protobuf.FloatValue": google_protobuf.FloatValue,
    ".google.protobuf.Int32Value": google_protobuf.Int32Value,
    ".google.protobuf.Int64Value": google_protobuf.Int64Value,
    ".google.protobuf.UInt32Value": google_protobuf.UInt32Value,
    ".google.protobuf.UInt64Value": google_protobuf.UInt64Value,
    ".google.protobuf.BoolValue": google_protobuf.BoolValue,
    ".google.protobuf.StringValue": google_protobuf.StringValue,
    ".google.protobuf.BytesValue": google_protobuf.BytesValue,
}


def parse_source_type_name(field_type_name: str, request: PluginRequestCompiler) -> tuple[str, str]:
    """
    Split full source type name into package and type name.
    E.g. 'root.package.Message' -> ('root.package', 'Message')
         'root.Message.SomeEnum' -> ('root', 'Message.SomeEnum')

    The function goes through the symbols that have been defined (names, enums,
    packages) to find the actual package and name of the object that is referenced.
    """
    if field_type_name[0] != ".":
        raise RuntimeError("relative names are not supported")
    field_type_name = field_type_name[1:]
    parts = field_type_name.split(".")

    answer = None

    # a.b.c:
    # i=0: "", "a.b.c"
    # i=1: "a", "b.c"
    # i=2: "a.b", "c"
    for i in range(len(parts)):
        package_name, object_name = ".".join(parts[:i]), ".".join(parts[i:])

        package = request.output_packages.get(package_name)

        if not package:
            continue

        if object_name in package.messages or object_name in package.enums:
            if answer:
                # This should have already been handeled by protoc
                raise ValueError(f"ambiguous definition: {field_type_name}")
            answer = package_name, object_name

    if answer:
        return answer

    raise ValueError(f"can't find type name: {field_type_name}")


def get_type_reference(
    *,
    package: str,
    imports: set,
    source_type: str,
    request: PluginRequestCompiler,
    unwrap: bool = True,
    settings: Settings,
) -> str:
    """
    Return a Python type name for a proto type reference. Adds the import if
    necessary. Unwraps well known type if required.
    """
    if unwrap:
        if source_type in WRAPPER_TYPES:
            wrapped_type = type(WRAPPER_TYPES[source_type]().value)
            return f"{wrapped_type.__name__} | None"

        if source_type == ".google.protobuf.Duration":
            return "datetime.timedelta"

        elif source_type == ".google.protobuf.Timestamp":
            return "datetime.datetime"

    source_package, source_type = parse_source_type_name(source_type, request)

    current_package: list[str] = package.split(".") if package else []
    py_package: list[str] = source_package.split(".") if source_package else []
    py_type: str = pythonize_class_name(source_type)

    if py_package == current_package:
        return reference_sibling(py_type)

    if py_package[: len(current_package)] == current_package:
        return reference_descendent(current_package, imports, py_package, py_type)

    if current_package[: len(py_package)] == py_package:
        return reference_ancestor(current_package, imports, py_package, py_type)

    return reference_cousin(current_package, imports, py_package, py_type)


def reference_absolute(imports: set[str], py_package: list[str], py_type: str) -> str:
    """
    Returns a reference to a python type located in the root, i.e. sys.path.
    """
    string_import = ".".join(py_package)
    string_alias = "__".join([safe_snake_case(name) for name in py_package])
    imports.add(f"import {string_import} as {string_alias}")
    return f"{string_alias}.{py_type}"


def reference_sibling(py_type: str) -> str:
    """
    Returns a reference to a python type within the same package as the current package.
    """
    return f"{py_type}"


def reference_descendent(current_package: list[str], imports: set[str], py_package: list[str], py_type: str) -> str:
    """
    Returns a reference to a python type in a package that is a descendent of the
    current package, and adds the required import that is aliased to avoid name
    conflicts.
    """
    importing_descendent = py_package[len(current_package) :]
    string_from = ".".join(importing_descendent[:-1])
    string_import = importing_descendent[-1]
    if string_from:
        string_alias = "_".join(importing_descendent)
        imports.add(f"from .{string_from} import {string_import} as {string_alias}")
        return f"{string_alias}.{py_type}"
    else:
        imports.add(f"from . import {string_import}")
        return f"{string_import}.{py_type}"


def reference_ancestor(current_package: list[str], imports: set[str], py_package: list[str], py_type: str) -> str:
    """
    Returns a reference to a python type in a package which is an ancestor to the
    current package, and adds the required import that is aliased (if possible) to avoid
    name conflicts.

    Adds trailing __ to avoid name mangling (python.org/dev/peps/pep-0008/#id34).
    """
    distance_up = len(current_package) - len(py_package)
    if py_package:
        string_import = py_package[-1]
        string_alias = f"_{'_' * distance_up}{string_import}__"
        string_from = f"..{'.' * distance_up}"
        imports.add(f"from {string_from} import {string_import} as {string_alias}")
        return f"{string_alias}.{py_type}"
    else:
        string_alias = f"{'_' * distance_up}{py_type}__"
        imports.add(f"from .{'.' * distance_up} import {py_type} as {string_alias}")
        return string_alias


def reference_cousin(current_package: list[str], imports: set[str], py_package: list[str], py_type: str) -> str:
    """
    Returns a reference to a python type in a package that is not descendent, ancestor
    or sibling, and adds the required import that is aliased to avoid name conflicts.
    """
    shared_ancestry = os.path.commonprefix([current_package, py_package])  # type: ignore
    distance_up = len(current_package) - len(shared_ancestry)
    string_from = f".{'.' * distance_up}" + ".".join(py_package[len(shared_ancestry) : -1])
    string_import = py_package[-1]
    # Add trailing __ to avoid name mangling (python.org/dev/peps/pep-0008/#id34)
    # string_alias = f"{'_' * distance_up}" + safe_snake_case(".".join(py_package[len(shared_ancestry) :])) + "__"
    string_alias = (
        f"{'_' * distance_up}"
        + "__".join([safe_snake_case(name) for name in py_package[len(shared_ancestry) :]])
        + "__"
    )
    imports.add(f"from {string_from} import {string_import} as {string_alias}")
    return f"{string_alias}.{py_type}"



from collections.abc import Callable

from .any import Any
from .duration import Duration
from .timestamp import Timestamp

# For each (package, message name), lists the methods that should be added to the message definition.
# The source code of the method is read from the `known_types` folder. If imports are needed, they can be directly added
# to the template file: they will automatically be removed if not necessary.
KNOWN_METHODS: dict[tuple[str, str], list[Callable]] = {
    ("google.protobuf", "Any"): [Any.pack, Any.unpack, Any.to_dict],
    ("google.protobuf", "Timestamp"): [Timestamp.from_datetime, Timestamp.to_datetime, Timestamp.timestamp_to_json],
    ("google.protobuf", "Duration"): [Duration.from_timedelta, Duration.to_timedelta, Duration.delta_to_json],
}
